#include <windows.h>
#include <gl\gl.h>
#include <vector>
#include <iostream>
#include "tpl.h"
#include "common.h"
#include "color.h"

using namespace std;

TPL::TPL ()
{
}

TPL::~TPL ()
{
}

bool TPL::LoadFile (const string &filename)
{
	m_filename = filename + '-';
	
	FILE* fh;
	fh = fopen(m_filename.c_str(), "rb");
	if (fh == NULL)
		return false;

	// Read Header
	fread(&m_header, sizeof(TPLHeader), 1, fh);
	byteSwapU32(m_header.magic);
	byteSwapU32(m_header.ntextures);
	byteSwapU32(m_header.sizeHeader);

	// Read Textures
	m_textures.resize(m_header.ntextures);
	fread(&m_textures[0], sizeof(TPLTexture), m_header.ntextures, fh);
	for (int x=0; x<m_textures.size(); x++)
	{
		byteSwapU32(m_textures[x].texHeaderOffset);
		byteSwapU32(m_textures[x].texPalOffset);
	}

	// Read Tex Headers
	m_texHeaders.resize(m_header.ntextures);
	for (int x=0; x<m_textures.size(); x++)
	{
		fseek(fh, m_textures[x].texHeaderOffset, SEEK_SET);
		fread(&m_texHeaders[x], sizeof(TPLTexHeader), 1, fh);
		
		byteSwapU16(m_texHeaders[x].height);
		byteSwapU16(m_texHeaders[x].width);
		byteSwapU32(m_texHeaders[x].format);
		byteSwapU32(m_texHeaders[x].texDataOffset);
		byteSwapU32(m_texHeaders[x].wrap_s);
		byteSwapU32(m_texHeaders[x].wrap_t);
		byteSwapU32(m_texHeaders[x].min_filter);
		byteSwapU32(m_texHeaders[x].mag_filter);
		byteSwapF32(m_texHeaders[x].lod_bias);
	}

	fseek(fh, 0, SEEK_END);
	u32 filesize = ftell(fh);

	// Read Data
	m_tex.resize(m_header.ntextures);
	for (int x=0; x<m_tex.size(); x++)
	{
		m_tex[x].texHeader = &m_texHeaders[x];

		// Calculate Size
		u32 length = 0;
		if (x < m_tex.size()-1)
			length = m_texHeaders[x+1].texDataOffset;
		else
			length = filesize;
		length -= m_tex[x].texHeader->texDataOffset;

		m_tex[x].data.resize(length);
		fseek(fh, m_tex[x].texHeader->texDataOffset, SEEK_SET);
		fread(&m_tex[x].data[0], sizeof(u8), length, fh);
	}

	fclose(fh);
	
	return true;
}

vector<Color> TPL::DecodeTexture (u32 index)
{
	switch (m_tex[index].texHeader->format) {
		case 0: return DecodeStandard(index, 0);
		case 1: return DecodeStandard(index, 1);
		case 2: return DecodeStandard(index, 2);
		case 3: return DecodeStandard(index, 3);
		case 4: return DecodeStandard(index, 4);
		case 5: return DecodeStandard(index, 5);
		case 6: return DecodeStandard(index, 6);
		case 14: return DecodeS3TC(index);
		default:
		{
			vector<Color> data;
			data.resize(m_tex[index].texHeader->width * m_tex[index].texHeader->height);
			return data;
		}
	}
}

vector<Color> TPL::DecodeS3TC (u32 index)
{
	TPLItem ti = m_tex[index];
	TPLTexHeader* th = m_tex[index].texHeader;

	int width = th->width;
	int tilesWide = th->width / 8;
	int tilesHigh = th->height / 8;

	vector<Color> data;
	data.resize(th->height*th->width);

	// Texture level
	for (int y=0; y<tilesHigh; y++) {
		for (int x=0; x<tilesWide; x++)	{
			int texAddr = y*tilesWide*4*8 + x*4*8;

			// Tile Level
			for (int z=0; z<4; z++)	{
				int bTexAddr = texAddr + z*8;

				// Block Level
				int pxb = x*8;
				int pyb = y*8;
				switch (z) {
					case 0: break;
					case 1: pxb += 4; break;
					case 2: pyb += 4; break;
					case 3: pxb += 4; pyb += 4; break;
				}
				
				// Generate Colors
				Color c0, c1, c2, c3;
				c0.RGB565_Unpack((ti.data[bTexAddr]<<8) | ti.data[bTexAddr+1]);
				c1.RGB565_Unpack((ti.data[bTexAddr+2]<<8) | ti.data[bTexAddr+3]);
				if (c0.RGBA8_Get() > c1.RGBA8_Get()) {
					c2.Blend(c0, c1, 85);
					c3.Blend(c0, c1, 171);
				} else {
					c2.Blend(c0, c1, 128);
					c3.Set(0, 0, 0, 0);
				}

				// Generate Texels
				int texels[4][4];
				for (int a=0; a<4; a++)	{
					texels[a][0] = (ti.data[bTexAddr+4+a] & 0xC0) >> 6;
					texels[a][1] = (ti.data[bTexAddr+4+a] & 0x30) >> 4;
					texels[a][2] = (ti.data[bTexAddr+4+a] & 0x0C) >> 2;
					texels[a][3] = (ti.data[bTexAddr+4+a] & 0x03) >> 0;
				}

				// Map Colors
				for (int a=0; a<4; a++) {
					for (int b=0; b<4; b++) {
						int px = pxb + b;
						int py = pyb + a;

						//if (py*width + px < th->height*th->width)
						switch (texels[a][b]) {
							case 0: data[py*width + px] = c0; break;
							case 1: data[py*width + px] = c1; break;
							case 2: data[py*width + px] = c2; break;
							case 3: data[py*width + px] = c3; break;
						}
					}
				}
			}
		}
	}

	return data;
}

vector<Color> TPL::DecodeStandard (u32 index, u32 type)
{
	TPLItem ti = m_tex[index];
	TPLTexHeader th = *m_tex[index].texHeader;

	u32 tileWidth = getTileWidth(type);
	u32 tileHeight = getTileHeight(type);
	u32 tilesWide = th.width / tileWidth;
	u32 tilesHigh = th.height / tileHeight;
	
	f32 step = getBytePP(type);
	u32 tsize = tileWidth * tileHeight * step;
	
	vector<Color> data;
	data.resize(th.width*th.height);

	for (u32 y=0; y<tilesHigh; y++) {
		for (u32 x=0; x<tilesWide; x++) {
			u32 addr = y*tilesWide*tsize + x*tsize;
			u32 pyb = y*tileHeight;
			u32 pxb = x*tileWidth;
			for (u32 a=0; a<tileHeight; a++) {
				for (u32 b=0; b<tileWidth; b++) {
					u32 baddr = addr + a*tileWidth*step + b*step;
					u32 py = pyb + a;
					u32 px = pxb + b;

					Color c0;
					switch (type) {
						case 0: c0.I4_Unpack(ti.data[baddr] >> (4 * ((b+1)%2))); break;
						case 1: c0.I8_Unpack(ti.data[baddr]); break;
						case 2: c0.IA4_Unpack(ti.data[baddr]); break;
						case 3: c0.IA8_Unpack((ti.data[baddr] << 8) | ti.data[baddr+1]); break;
						case 4: c0.RGB565_Unpack((ti.data[baddr] << 8) | ti.data[baddr+1]); break;
						case 5: c0.RGB5A3_Unpack((ti.data[baddr] << 8) | ti.data[baddr+1]); break;
						case 6: c0.RGBA8_Unpack((ti.data[baddr] << 8) | ti.data[baddr+1],
												(ti.data[baddr+32] << 8) | ti.data[baddr+31]); break;
					}
					data[py*th.width+px] = c0;
				}
			}
		}
	}

	return data;
}

u32 TPL::getTileWidth (u32 type)
{
	switch (type)
	{
	case 0: return 8;
	case 1: return 8;
	case 2: return 8;
	case 3: return 4;
	case 4: return 4;
	case 5: return 4;
	case 6: return 4;
	case 8: return 8;
	case 9: return 8;
	case 10: return 4;
	case 14: return 4;
	default: return 0;
	}
}

u32 TPL::getTileHeight (u32 type)
{
	switch (type)
	{
	case 0: return 8;
	case 1: return 4;
	case 2: return 4;
	case 3: return 4;
	case 4: return 4;
	case 5: return 4;
	case 6: return 4;
	case 8: return 8;
	case 9: return 4;
	case 10: return 4;
	case 14: return 4;
	default: return 0;
	}
}

f32 TPL::getBytePP (u32 type)
{
	switch (type)
	{
	case 0: return 0.5f;
	case 1: return 1.0f;
	case 2: return 1.0f;
	case 3: return 2.0f;
	case 4: return 2.0f;
	case 5: return 2.0f;
	case 6: return 4.0f;
	case 8: return 0.5f;
	case 9: return 1.0f;
	case 10: return 2.0f;
	case 14: return 0.25f;
	default: return 0.0f;
	}
}